@@ -14,6 +14,7 @@ import subprocess
 import sys
 import time
 import traceback
+from copy import deepcopy
 from dataclasses import dataclass, replace
 from multiprocessing import cpu_count, Pool
 from pathlib import Path
@@ -21,6 +22,7 @@ from typing import Any, AnyStr, cast, Dict, List, Optional, Sequence, Union
 
 from libcst import parse_module, PartialParserConfig
 from libcst.codemod._codemod import Codemod
+from libcst.codemod._context import CodemodContext
 from libcst.codemod._dummy_pool import DummyPool
 from libcst.codemod._runner import (
     SkipFile,
@@ -90,7 +92,10 @@ def gather_files(
             ret.extend(
                 str(p)
                 for p in Path(fd).rglob("*.py*")
-                if str(p).endswith("py") or (include_stubs and str(p).endswith("pyi"))
+                if Path.is_file(p)
+                and (
+                    str(p).endswith("py") or (include_stubs and str(p).endswith("pyi"))
+                )
             )
     return sorted(ret)
 
@@ -211,6 +216,7 @@ def _execute_transform(  # noqa: C901
     transformer: Codemod,
     filename: str,
     config: ExecutionConfig,
+    scratch: Dict[str, object],
 ) -> ExecutionResult:
     for pattern in config.blacklist_patterns:
         if re.fullmatch(pattern, filename):
@@ -241,30 +247,28 @@ def _execute_transform(  # noqa: C901
                 ),
             )
 
-        # Somewhat gross hack to provide the filename in the transform's context.
-        # We do this after the fork so that a context that was initialized with
-        # some defaults before calling parallel_exec_transform_with_prettyprint
-        # will be updated per-file.
-        transformer.context = replace(
-            transformer.context,
-            filename=filename,
-            scratch={},
-        )
-
         # determine the module and package name for this file
         try:
             module_name_and_package = calculate_module_and_package(
                 config.repo_root or ".", filename
             )
-            transformer.context = replace(
-                transformer.context,
-                full_module_name=module_name_and_package.name,
-                full_package_name=module_name_and_package.package,
-            )
+            mod_name = module_name_and_package.name
+            pkg_name = module_name_and_package.package
         except ValueError as ex:
             print(
                 f"Failed to determine module name for {filename}: {ex}", file=sys.stderr
             )
+            mod_name = None
+            pkg_name = None
+
+        # Apart from metadata_manager, every field of context should be reset per file
+        transformer.context = CodemodContext(
+            scratch=deepcopy(scratch),
+            filename=filename,
+            full_module_name=mod_name,
+            full_package_name=pkg_name,
+            metadata_manager=transformer.context.metadata_manager,
+        )
 
         # Run the transform, bail if we failed or if we aren't formatting code
         try:
@@ -415,7 +419,7 @@ class Progress:
         operations still to do.
         """
 
-        if files_finished <= 0:
+        if files_finished <= 0 or elapsed_seconds == 0:
             # Technically infinite but calculating sounds better.
             return "[calculating]"
 
@@ -473,7 +477,7 @@ def _print_parallel_result(
             )
 
         # In unified diff mode, the code is a diff we must print.
-        if unified_diff:
+        if unified_diff and result.code:
             print(result.code)
 
 
@@ -557,7 +561,7 @@ def parallel_exec_transform_with_prettyprint(  # noqa: C901
 
     # Ensure that we have no duplicates, otherwise we might get race conditions
     # on write.
-    files = sorted(list({os.path.abspath(f) for f in files}))
+    files = sorted({os.path.abspath(f) for f in files})
     total = len(files)
     progress = Progress(enabled=not hide_progress, total=total)
 
@@ -631,6 +635,7 @@ def parallel_exec_transform_with_prettyprint(  # noqa: C901
                 "transformer": transform,
                 "filename": filename,
                 "config": config,
+                "scratch": transform.context.scratch,
             }
             for filename in files
         ]
